#!/usr/bin/env python
'''Python3.4 unit testing framework for FW4EX based on unittest.

This framework brings new features. (1) successful assertions are
counted. (2) numbers of successful tests, failures, assertions is
printed in a bash-friendly format. (3) partial results are flushed in
a file as soon as known (this allows the tests to be killed even if
unfinished and still get the latest results).

Simple usage:
   
     import fw4exunittest

     class MyTestCase(fw4exunittest.TestCase):
       def test_1(self):
         # almost similar code as in unittest.TestCase except that, in 
         # order to not reveal the values that are expected in assertions, 
         # you should not write:
         assertEqual(computation(1,2,3), 42

         # but rather:
         expected = 42
         assertEqual(computation(1,2,3), expected)

         # This is so because, for a failing assertion, the last line of the
         # text of the assertion is printed and knowing what value to return
         # allows student to easily cheat!

         # Remember also that a testXX method stops when an assertion is not
         # verified but other tests will be executed. 
         # Remember also that testXX methods are ran in alphabetic order.

     if __name__ == '__main__':
         fw4exunittest.main()

Start the tests with:

     python test_something.py --result /tmp/lr.sh

and watch the results in /tmp/lr.sh (default is ./lastResults.sh).
The unittest module will fill stdout and stderr as usual except that
some parts of the traceback section are removed in order to prevent
leaking the values that are expected in the tests.

'''

__author__ = "Christian Queinnec"
__email__ = "christian dot queinnec at paracamplus dot com"
__version__ = "0.1"

import unittest
import argparse
import warnings
import time
import traceback
import re

class TextTestResult(unittest.TextTestResult):
    """Holder for test result information.
     
    Just add to the parent unittest.TestResult
       - an attribute to record the number of attempted assertions
       - a method to stringify the results in a shell friendly way.

    CAUTION: this class needs its 'resultfile' attribute. This attribute 
    is set by TextTestRunner._makeResult()

    """
    def __init__(self, stream=None, descriptions=None, verbosity=1):
        unittest.TextTestResult.__init__(self, stream, descriptions, verbosity)
        self.assertions = 0
        self.successful_assertions = 0

    def __repr__(self):
        s = "<%s run=%i errors=%i failures=%i assertions=%i/%i resultfile=%s>" % \
         (str(self.__class__), self.testsRun, len(self.errors),
          len(self.failures), 
          self.successful_assertions, 
          self.assertions, 
          self.resultfile)
        return s

    def generateResults(self):
        s = "#/bin/bash\n"
        s += "#number of TESTS attempted:\n"
        s += "TESTS=%d\n" % self.testsRun
        failed, errored = map(len, (self.failures, self.errors))
        s += "#number of failed TESTS:\n"
        s += "FAILURES=%d\n" % (failed + errored)
        # errored ???
        s += "#number of attempted assertions:\n" 
        s += "ATTEMPTED_ASSERTIONS=%d\n" % self.assertions
        s += "#number of successful assertions:\n"
        s += "ASSERTIONS=%d\n" % self.successful_assertions
        return s

    def writeResults(self):
        #import sys; sys.stdout.write(str(self) + "\n")   # DEBUG
        s = self.generateResults()
        with open(self.resultfile, "w") as out:
            out.write(s)

    def _exc_info_to_string(self, err, test):
        """Converts a sys.exc_info()-style tuple of values into a string."""
        exctype, value, tb = err
        # Skip test runner traceback levels
        while tb and self._is_relevant_tb_level(tb):
            tb = tb.tb_next

        if exctype is test.failureException:
            # Skip assert*() traceback levels
            length = self._count_relevant_tb_levels(tb)
            msgLines = traceback.format_exception(exctype, value, tb, length)
        else:
            msgLines = traceback.format_exception(exctype, value, tb)
        
        # remove information revealing values expected from assertions:
        keptLines = []
        for item in msgLines:
            if re.match('^Assertion.*', item):
                pass
            elif re.search('/fw4exunittest.py', item):
                break
            else:
                keptLines.append(item)
        msgLines = keptLines

        if self.buffer:
            output = sys.stdout.getvalue()
            error = sys.stderr.getvalue()
            if output:
                if not output.endswith('\n'):
                    output += '\n'
                msgLines.append(STDOUT_LINE % output)
            if error:
                if not error.endswith('\n'):
                    error += '\n'
                msgLines.append(STDERR_LINE % error)
        return ''.join(msgLines)


class TextTestRunner(unittest.TextTestRunner):
    """A test runner class that displays results in textual form.

    """
    def __init__(self, stream=None, 
                 description=True, verbosity=1,
                 failfast=False, buffer=False, 
                 resultclass=TextTestResult, warnings=None):
        unittest.TextTestRunner.__init__(
            self, stream, description, verbosity,
            failfast, buffer, resultclass, warnings)
        self.resultclass = TextTestResult
        self.result = None

    def _makeResult(self):
        if self.result is None:
            result = self.resultclass(self.stream, 
                                      self.descriptions, 
                                      self.verbosity)
            result.resultfile = self.resultfile
            self.result = result
        return self.result

    def run(self, test):
        "Run the given test case or test suite."
        # (1) create the result now (rather than in super().run(test)):
        self.result = self._makeResult()

        # (2) Patch all TestCases so they know this result
        # Here, test is a TestSuite of TestCases or TestSuites
        def fix(item):
            if isinstance(item, TestCase):
                item.testresult = self.result
            elif isinstance(item, unittest.TestSuite):
                for index, subitem in enumerate(item):
                    fix(subitem)
        fix(test)
        
        # (3) run as before:
        self.result = super().run(test);
        self.result.writeResults()
        return self.result

# NOTA: TestCase needs a TextTestResult to record the number of
# assertions, while TextTestRunner needs to read that TextTestResult
# to generate the final results of all tests. It must be the same object.

class TestCase(unittest.TestCase):
    """An extended TestCase class.

    CAUTION: this class needs a 'testresult' attribute holding the
    TestResult object (holding the 'assertions' and
    'successful_assertions' attributes to increment).

    """
    def __init__(self, methodName='runTest'):
        self.testresult = None
        unittest.TestCase.__init__(self, methodName)

    def recordAssertions(self):
        if self.testresult is not None:
            # Count all assertions:
            self.testresult.assertions += 1

    def recordResults(self):
        if self.testresult is not None:
            # Count successful assertions:
            self.testresult.successful_assertions += 1
            self.testresult.writeResults()

    def FW4EXassert(assertName):
        def wrap(f):
            def wrapped(self, *args):
                self.recordAssertions()
                f(self, *args)
                self.recordResults()
            return wrapped
        return wrap(unittest.TestCase.__dict__[assertName])

    assertFalse = FW4EXassert('assertFalse')
    assertTrue  = FW4EXassert('assertTrue')
    assertRaises = FW4EXassert('assertRaises')
    assertWarns = FW4EXassert('assertWarns')
    assertEqual = FW4EXassert('assertEqual')
    assertNotEqual = FW4EXassert('assertNotEqual')
    assertAlmostEqual = FW4EXassert('assertAlmostEqual')
    assertNotAlmostEqual = FW4EXassert('assertNotAlmostEqual')
    assertSequenceEqual = FW4EXassert('assertSequenceEqual')
    assertListEqual = FW4EXassert('assertListEqual')
    assertTupleEqual = FW4EXassert('assertTupleEqual')
    assertSetEqual = FW4EXassert('assertSetEqual')
    assertIn = FW4EXassert('assertIn')
    assertNotIn = FW4EXassert('assertNotIn')
    assertIs = FW4EXassert('assertIs')
    assertIsNot = FW4EXassert('assertIsNot')
    assertDictEqual = FW4EXassert('assertDictEqual')
    assertDictContainsSubset = FW4EXassert('assertDictContainsSubset')
    assertCountEqual = FW4EXassert('assertCountEqual')
    assertMultiLineEqual = FW4EXassert('assertMultiLineEqual')
    assertLess = FW4EXassert('assertLess')
    assertLessEqual = FW4EXassert('assertLessEqual')
    assertGreater = FW4EXassert('assertGreater')
    assertGreaterEqual = FW4EXassert('assertGreaterEqual')
    assertIsNone = FW4EXassert('assertIsNone')
    assertIsNotNone = FW4EXassert('assertIsNotNone')
    assertIsInstance = FW4EXassert('assertIsInstance')
    assertNotIsInstance = FW4EXassert('assertNotIsInstance')
    assertRaisesRegex = FW4EXassert('assertRaisesRegex')
    assertWarnsRegex = FW4EXassert('assertWarnsRegex')
    assertRegex = FW4EXassert('assertRegex')
    assertNotRegex = FW4EXassert('assertNotRegex')

class TestProgram(unittest.TestProgram):
    """Extended command-line program.

This command-line program adds a new switch:
    -r, --resultfile     file where to store tests results
where the bash-friendly result of the tests will be stored after every
assertion check.
    """
    def runTests(self):
        #import sys; sys.stdout.write(str(self))  # DEBUG
        self.testRunner = TextTestRunner(verbosity=self.verbosity,
                                         failfast=self.failfast,
                                         buffer=self.buffer,
                                         warnings=self.warnings)
        self.testRunner.resultfile = self.resultfile
        # This will trigger self.testRunner.run(self):
        unittest.TestProgram.runTests(self)

    def _getParentArgParser(self):
        parser = super()._getParentArgParser()

        parser.add_argument('-r', '--resultfile',
                            default='lastResults.sh',
                            help='Filename where to store partial results')
        return parser

    def __repr__(self):
        return ("<%s resultfile=%s>" %
               (str(self.__class__), self.resultfile))

main = TestProgram

# end.
